import csv
import time
from operator import itemgetter


from helpers.data_set import DataSet
from helpers.metrics import *
from helpers.write_results import *
from trans_model import TransitionModel
from structure import ModelStructure


def experiment(ds_path, frequency_matrix_path, dist_func, threshold, test_proportion, prefix_len, debug_plot,
               prediction_len=-1):
    """Conduct measurements for Transitional model
    :param ds_path: path to data set
    :param frequency_matrix_path: path to frequency transition matrix
    :param dist_func: evaluation metric for finding similar instances by prefix
    :param threshold: similarity threshold
    :param test_proportion: fraction of test sample examples from whole data set
    :param prefix_len: length of prefix for prediction
    :param prediction_len: number of next predicted events after the prefix
    :param debug_plot: whether to plot or not graph structures
    :return: list of accuracy metrics for test samples
    """
    ds = DataSet(data_set_path=ds_path)
    train_set, test_set = ds.split_data(test_proportion)
    structure = ModelStructure(freq_matrix_path=frequency_matrix_path)
    digraph = structure.set_up()
    model = TransitionModel(train_set=train_set, digraph=digraph)
    results = []
    for case in test_set:
        # filter too short instances
        if len(case) < (prefix_len + prediction_len):
            continue
        prefix = case[:prefix_len]
        true_value = case[prefix_len:]
        prediction = model.predict(prefix, dist_func, threshold, plot=debug_plot)
        if not prediction:
            continue
        if prediction_len == -1:
            prediction_len = len(true_value)
        res = accuracy(first=prediction[:prediction_len], second=true_value[:prediction_len])
        res = [res[0], res[1], res[2]] + true_value[:prediction_len] + ["*"] + prediction[:prediction_len]
        results.append(res)
    return results


def cv_experiment(ds_path, frequency_matrix_path, dist_func, threshold, n_folds, prefix_len, debug_plot,
                  prediction_len=-1):
    """Conduct n-fold cross-validation for Transitional model
    :param ds_path: path to data set
    :param frequency_matrix_path: path to frequency transition matrix
    :param dist_func: evaluation metric for finding similar instances by prefix
    :param threshold: similarity threshold
    :param test_proportion: fraction of test sample examples from whole data set
    :param prefix_len: length of prefix for prediction
    :param prediction_len: number of next predicted events after the prefix
    :param debug_plot: whether to plot or not graph structures
    :return: list of accuracy metrics for test samples
    """
    ds = DataSet(data_set_path=ds_path)
    label_kfold = ds.split_k_fold(n_folds)
    results = []
    for ind, (train_index, test_idndex) in enumerate(label_kfold, 1):
        print("fold #{:d}".format(ind))
        train_set = itemgetter(*train_index)(ds.data_set)
        test_set = itemgetter(*test_idndex)(ds.data_set)
        structure = ModelStructure(freq_matrix_path=frequency_matrix_path)
        digraph = structure.set_up()
        model = TransitionModel(train_set=train_set, digraph=digraph)
        current_result = []
        for case in test_set:
            # filter too short instances
            if len(case) < (prefix_len + prediction_len):
                continue
            prefix = case[:prefix_len]
            true_value = case[prefix_len:]
            prediction = model.predict(prefix, dist_func, threshold, plot=debug_plot)
            if not prediction:
                continue
            if prediction_len == -1:
                prediction_len = len(true_value)
            res = accuracy(first=prediction[:prediction_len], second=true_value[:prediction_len])
            res = [res[0], res[1], res[2]] + true_value[:prediction_len] + ["*"] + prediction[:prediction_len]
            current_result.append(res)
        results.append(current_result)
    return results


def read_experiment_parameters(file_path):
    """Read parameters for conducting automated experiment
     :param file_path: path to file with parameters
     :return: list of parameter's sets
    """
    param_reader = csv.reader(open(file_path))
    rows_in_file = []
    for row in param_reader:
        rows_in_file.append(row)

    param_set = []
    for num, row in enumerate(rows_in_file[1:], 1):
        file_name = row[0]
        method = row[1]
        if method == 'lv':
            dist_func=is_similar_by_levenstein
        elif method == 'dl':
            dist_func=is_similar_by_damerau_levenstein
        else:
            print("Unidentified method!", "Skip parameters set #{:d}".format(num))
            continue
        threshold = float(row[2])
        test_proportion = float(row[3])
        prefixes_len = [int(x) for x in row[4].split(";")]
        prediction_len = [int(x) for x in row[5].split(";")]
        param_set.append([file_name, dist_func, threshold, test_proportion, prefixes_len, prediction_len, method])
    return param_set


def measurements(plan_file, results_dir="evaluation/results", data_dir="evaluation/",
                 freq_matr_dir="evaluation/tr_system/", plot=False, cv=True, folds=3):
    """Conduct measurements for all parameters
    :param plan_file: path to file with parameters
    :param results_dir: path to result directory
    :param data_dir: path to directory with data
    :param freq_matr_dir: path to frequency/adjacency matrix
    :param plot: whether to plot or not graph structures
    :param cv: whether to make cross-validation or usual experiment
    :param folds: number of folds for cross-validation
    """
    param_sets = read_experiment_parameters(plan_file)
    for ind, param in enumerate(param_sets, 1):
        start = time.clock()
        base_file_name = param[0]
        dist_func = param[1]
        threshold = param[2]
        test_proportion = param[3]
        prefix_lens = param[4]
        prediction_lens = param[5]
        print("Evaluation for {:s} file with {:s} metric and threshold {:.2f}".format(base_file_name, param[6],
                                                                                      threshold))
        try:
            for pref_len in prefix_lens:
                for pred_len in prediction_lens:
                    print("Prefix length: {:d}, Prediction length: {:d}".format(pref_len, pred_len))
                    if cv:
                        results = cv_experiment(ds_path=data_dir + base_file_name,
                                             frequency_matrix_path=freq_matr_dir + base_file_name,
                                             n_folds=folds, dist_func=dist_func, threshold=threshold,
                                             prefix_len=pref_len, prediction_len=pred_len, debug_plot=plot)
                        result_file = results_dir + "/folds{:d}_{:s}_{:s}_{:d}_{:d}_".format(folds, param[6],
                                                        str(param[2]), pref_len, pred_len) + base_file_name
                        if results:
                            write_cv_results_to(result_file, results)
                        else:
                            print("There is no results for cv!")
                    else:
                        results = experiment(ds_path=data_dir + base_file_name,
                                             frequency_matrix_path=freq_matr_dir + base_file_name,
                                             test_proportion=test_proportion, dist_func=dist_func,
                                             threshold=threshold, prefix_len=pref_len,
                                             prediction_len=pred_len, debug_plot=plot)
                        result_file = results_dir + "/{:s}_{:s}_{:d}_{:d}_".format(param[6], str(param[2]), pref_len,
                                                                                   pred_len) + base_file_name
                        if results:
                            write_result_to(result_file, results)
                        else:
                            print("There is no results")
        except Exception as error:
            print("Error was occurred. Skip processing this parameters:")
            print(error)
            continue
        finally:
            end = time.clock()
            print("Time elapsed: {:.3f}".format(end - start))
            print("*"*20)


if __name__ == "__main__":
    measurements(plan_file="evaluation/plan.csv", cv=True, plot=False, folds=3)
