import random
import csv
import time


from helpers.data_set import DataSet
from helpers.metrics import *
from helpers.write_results import *
from structure import ModelStructure


def random_prediction(matrix, start_node, prediction_len):
    """emulates random walk starting from provided event
    :param matrix: frequency/adjacency matrix
    :param start_node: starting event for prediction
    :param prediction_len: maximal number of events in random walk
    :return: predicted events
    """
    current_node = start_node
    prediction = []
    for _ in range(prediction_len):
        current_node = get_rnd_neighbor(matrix, current_node)
        if not current_node:
            print("Stopped prediction as reached end node.")
            break
        prediction.append(current_node)
    return prediction


def get_rnd_neighbor(matrix, node):
    """returns one of the adjacent events in transition system with equal probability
    :param matrix: frequency/adjacency matrix
    :param node: node for finding its neighbor
    :return: random neighbor of the given node
    """
    neighbors = []
    col_ind = matrix[0].index(node)
    for row_ind, row in enumerate(matrix[1:], 1):
        if row[col_ind] != '0':
            neighbor = matrix[row_ind][0]
            neighbors.append(neighbor)
    if not neighbors:
        return None
    return random.choice(neighbors)


def experiment(ds_path, frequency_matrix_path, test_proportion, prefix_len, prediction_len=-1):
    """Conduct measurements for one Random Walk
    :param ds_path: path to data set
    :param frequency_matrix_path: path to frequency transition (adjacency) matrix
    :param test_proportion: fraction of test sample examples from whole data set
    :param prefix_len: length of prefix for prediction
    :param prediction_len: number of events in Random Walk
    :return: list of accuracy metrics for sequence events from Random Walk
    """
    ds = DataSet(data_set_path=ds_path)
    train_set, test_set = ds.split_data(test_proportion)
    matrix = ModelStructure(freq_matrix_path=frequency_matrix_path).freq_matrix
    results = []
    for indx, case in enumerate(test_set):
        # filter too short instances
        if len(case) < (prefix_len + prediction_len):
            continue
        true_value = case[prefix_len:]
        if prediction_len == -1:
            prediction_len = len(true_value)

        prediction = random_prediction(matrix, case[prefix_len-1], prediction_len)
        if not prediction:
            continue
        res = accuracy(first=prediction[:prediction_len], second=true_value[:prediction_len])
        res = [res[0], res[1], res[2]] + true_value[:prediction_len] + ["*"] + prediction[:prediction_len]
        results.append(res)
    return results


def read_parameters(file_path):
    """Read parameters for conducting automated experiment
    :param file_path: path to file with parameters
    :return: list of parameter's sets
    """
    param_reader = csv.reader(open(file_path))
    rows_in_file = []
    for row in param_reader:
        rows_in_file.append(row)

    param_set = []
    for num, row in enumerate(rows_in_file[1:], 1):
        file_name = row[0]
        prefixes_len = [int(x) for x in row[1].split(";")]
        prediction_len = [int(x) for x in row[2].split(";")]
        param_set.append([file_name, prefixes_len, prediction_len])
    return param_set


def measurements(plan_file, results_dir="../results/Random_model", data_dir="../data/",
                 freq_matr_dir="../data/case_freqs/"):
    """Conduct measurements for all parameters for random model
    :param plan_file: path to file with parameters
    :param results_dir: path to result directory
    :param freq_matr_dir: path to frequency/adjacency matrix
    """
    param_sets = read_parameters(plan_file)
    for ind, param in enumerate(param_sets, 1):
        start = time.clock()
        base_file_name = param[0]
        prefix_lens = param[1]
        prediction_lens = param[2]
        print("Evaluation for {:s} file".format(base_file_name))
        for pref_len in prefix_lens:
            for pred_len in prediction_lens:
                results = experiment(ds_path=data_dir + base_file_name,
                                     frequency_matrix_path=freq_matr_dir + base_file_name, test_proportion=0.2,
                                     prefix_len=pref_len, prediction_len=pred_len)
                result_file = results_dir + "/rnd_{:d}_{:d}_".format(pref_len, pred_len) + base_file_name
                if results:
                    write_result_to(result_file, results)
                else:
                    print("There is no results for {:d} param set".format(ind))
        end = time.clock()
        print(end - start)

if __name__ == "__main__":
    measurements(plan_file="../evaluation/plan_random.csv")
